@@ -1,11 +1,10 @@
 // SPDX-FileCopyrightText: 2021 Heiko Schaefer <heiko@schaefer.name>
 // SPDX-License-Identifier: MIT OR Apache-2.0
 
-//! This crate implements the experimental `ScdClient` backend for the
+//! This crate implements the experimental `ScdBackend`/`ScdTransaction` backend for the
 //! `openpgp-card` crate.
 //! It uses GnuPG's scdaemon (via GnuPG Agent) to access OpenPGP cards.
 
-use anyhow::{anyhow, Result};
 use futures::StreamExt;
 use lazy_static::lazy_static;
 use sequoia_ipc::assuan::Response;
@@ -13,12 +12,10 @@ use sequoia_ipc::gnupg::{Agent, Context};
 use std::sync::Mutex;
 use tokio::runtime::Runtime;
 
-use openpgp_card::{CardApp, Error};
-use openpgp_card::{CardCaps, CardClient};
+use openpgp_card::{CardBackend, CardCaps, CardTransaction, Error, PinType, SmartcardError};
 
 lazy_static! {
-    static ref RT: Mutex<Runtime> =
-        Mutex::new(tokio::runtime::Runtime::new().unwrap());
+    static ref RT: Mutex<Runtime> = Mutex::new(tokio::runtime::Runtime::new().unwrap());
 }
 
 /// The Assuan protocol (in GnuPG) limits the length of commands.
@@ -49,24 +46,23 @@ const ASSUAN_LINELENGTH: usize = 1000;
 /// In particular, uploading rsa4096 keys fails via scdaemon, with such cards.
 const APDU_CMD_BYTES_MAX: usize = (ASSUAN_LINELENGTH - 25) / 2;
 
-/// An implementation of the CardClient trait that uses GnuPG's scdaemon
+/// An implementation of the CardBackend trait that uses GnuPG's scdaemon
 /// (via GnuPG Agent) to access OpenPGP card devices.
-pub struct ScdClient {
+pub struct ScdBackend {
     agent: Agent,
     card_caps: Option<CardCaps>,
 }
 
-impl ScdClient {
+impl ScdBackend {
     /// Open a CardApp that uses an scdaemon instance as its backend.
     /// The specific card with AID `serial` is requested from scdaemon.
-    pub fn open_by_serial(
-        agent: Option<Agent>,
-        serial: &str,
-    ) -> Result<CardApp, Error> {
-        let mut card = ScdClient::new(agent, true)?;
+    pub fn open_by_serial(agent: Option<Agent>, serial: &str) -> Result<Self, Error> {
+        let mut card = ScdBackend::new(agent, true)?;
         card.select_card(serial)?;
 
-        Ok(CardApp::initialize(Box::new(card))?)
+        card.transaction()?.initialize()?;
+
+        Ok(card)
     }
 
     /// Open a CardApp that uses an scdaemon instance as its backend.
@@ -75,15 +71,17 @@ impl ScdClient {
     ///
     /// (NOTE: implicitly picking an unspecified card might be a bad idea.
     /// You might want to avoid using this function.)
-    pub fn open_yolo(agent: Option<Agent>) -> Result<CardApp, Error> {
-        let card = ScdClient::new(agent, true)?;
+    pub fn open_yolo(agent: Option<Agent>) -> Result<Self, Error> {
+        let mut card = ScdBackend::new(agent, true)?;
 
-        Ok(CardApp::initialize(Box::new(card))?)
+        card.transaction()?.initialize()?;
+
+        Ok(card)
     }
 
     /// Helper fn that shuts down scdaemon via GnuPG Agent.
     /// This may be useful to obtain access to a Smard card via PCSC.
-    pub fn shutdown_scd(agent: Option<Agent>) -> Result<()> {
+    pub fn shutdown_scd(agent: Option<Agent>) -> Result<(), Error> {
         let mut scdc = Self::new(agent, false)?;
 
         scdc.send("SCD RESTART")?;
@@ -97,13 +95,23 @@ impl ScdClient {
     ///
     /// If `agent` is None, a Context with the default GnuPG home directory
     /// is used.
-    fn new(agent: Option<Agent>, init: bool) -> Result<Self> {
+    fn new(agent: Option<Agent>, init: bool) -> Result<Self, Error> {
         let agent = if let Some(agent) = agent {
             agent
         } else {
             // Create and use a new Agent based on a default Context
-            let ctx = Context::new()?;
-            RT.lock().unwrap().block_on(Agent::connect(&ctx))?
+            let ctx = Context::new().map_err(|e| {
+                Error::Smartcard(SmartcardError::Error(format!("Context::new failed {}", e)))
+            })?;
+            RT.lock()
+                .unwrap()
+                .block_on(Agent::connect(&ctx))
+                .map_err(|e| {
+                    Error::Smartcard(SmartcardError::Error(format!(
+                        "Agent::connect failed {}",
+                        e
+                    )))
+                })?
         };
 
         let mut scdc = Self {
@@ -118,13 +126,22 @@ impl ScdClient {
         Ok(scdc)
     }
 
+    fn send2(&mut self, cmd: &str) -> Result<(), Error> {
+        self.agent.send(cmd).map_err(|e| {
+            Error::Smartcard(SmartcardError::Error(format!(
+                "scdc agent send failed: {}",
+                e
+            )))
+        })
+    }
+
     /// Call "SCD SERIALNO", which causes scdaemon to be started by gpg
     /// agent (if it's not running yet).
-    fn serialno(&mut self) -> Result<()> {
+    fn serialno(&mut self) -> Result<(), Error> {
         let mut rt = RT.lock().unwrap();
 
         let send = "SCD SERIALNO";
-        self.agent.send(send)?;
+        self.send2(send)?;
 
         while let Some(response) = rt.block_on(self.agent.next()) {
             log::debug!("init res: {:x?}", response);
@@ -139,14 +156,16 @@ impl ScdClient {
             }
         }
 
-        Err(anyhow!("SCDC init() failed"))
+        Err(Error::Smartcard(SmartcardError::Error(
+            "SCDC init() failed".into(),
+        )))
     }
 
     /// Ask scdameon to switch to using a specific OpenPGP card, based on
     /// its `serial`.
-    fn select_card(&mut self, serial: &str) -> Result<()> {
+    fn select_card(&mut self, serial: &str) -> Result<(), Error> {
         let send = format!("SCD SERIALNO --demand={}", serial);
-        self.agent.send(send)?;
+        self.send2(&send)?;
 
         let mut rt = RT.lock().unwrap();
 
@@ -154,7 +173,9 @@ impl ScdClient {
             log::debug!("select res: {:x?}", response);
 
             if response.is_err() {
-                return Err(anyhow!("Card not found"));
+                return Err(Error::Smartcard(SmartcardError::CardNotFound(
+                    serial.into(),
+                )));
             }
 
             if let Ok(Response::Status { .. }) = response {
@@ -167,11 +188,13 @@ impl ScdClient {
             }
         }
 
-        Err(anyhow!("Card not found"))
+        Err(Error::Smartcard(SmartcardError::CardNotFound(
+            serial.into(),
+        )))
     }
 
-    fn send(&mut self, cmd: &str) -> Result<()> {
-        self.agent.send(cmd)?;
+    fn send(&mut self, cmd: &str) -> Result<(), Error> {
+        self.send2(cmd)?;
 
         let mut rt = RT.lock().unwrap();
 
@@ -179,7 +202,7 @@ impl ScdClient {
             log::debug!("select res: {:x?}", response);
 
             if let Err(e) = response {
-                return Err(anyhow!("Err {:?}", e));
+                return Err(Error::Smartcard(SmartcardError::Error(format!("{:?}", e))));
             }
 
             if let Ok(..) = response {
@@ -192,23 +215,34 @@ impl ScdClient {
             }
         }
 
-        Err(anyhow!("Error sending command {}", cmd))
+        Err(Error::Smartcard(SmartcardError::Error(format!(
+            "Error sending command {}",
+            cmd
+        ))))
     }
 }
 
-impl CardClient for ScdClient {
+impl CardBackend for ScdBackend {
+    fn transaction(&mut self) -> Result<Box<dyn CardTransaction + Send + Sync + '_>, Error> {
+        Ok(Box::new(ScdTransaction { scd: self }))
+    }
+}
+
+pub struct ScdTransaction<'a> {
+    scd: &'a mut ScdBackend,
+}
+
+impl CardTransaction for ScdTransaction<'_> {
     fn transmit(&mut self, cmd: &[u8], _: usize) -> Result<Vec<u8>, Error> {
         log::trace!("SCDC cmd len {}", cmd.len());
 
         let hex = hex::encode(cmd);
 
         // (Unwrap is ok here, not having a card_caps is fine)
-        let ext = if self.card_caps.is_some()
-            && self.card_caps.unwrap().ext_support()
-        {
+        let ext = if self.card_caps().is_some() && self.card_caps().unwrap().ext_support() {
             // If we know about card_caps, and can do extended length we
             // set "exlen" accordingly ...
-            format!("--exlen={} ", self.card_caps.unwrap().max_rsp_bytes())
+            format!("--exlen={} ", self.card_caps().unwrap().max_rsp_bytes())
         } else {
             // ... otherwise don't send "exlen" to scdaemon
             "".to_string()
@@ -218,30 +252,30 @@ impl CardClient for ScdClient {
         log::debug!("SCDC command: '{}'", send);
 
         if send.len() > ASSUAN_LINELENGTH {
-            return Err(Error::InternalError(anyhow!(
+            return Err(Error::Smartcard(SmartcardError::Error(format!(
                 "APDU command is too long ({}) to send via Assuan",
                 send.len()
-            )));
+            ))));
         }
 
-        self.agent.send(send)?;
+        self.scd.send2(&send)?;
 
         let mut rt = RT.lock().unwrap();
 
-        while let Some(response) = rt.block_on(self.agent.next()) {
+        while let Some(response) = rt.block_on(self.scd.agent.next()) {
             log::debug!("res: {:x?}", response);
             if response.is_err() {
-                return Err(Error::InternalError(anyhow!(
+                return Err(Error::Smartcard(SmartcardError::Error(format!(
                     "Unexpected error response from SCD {:?}",
                     response
-                )));
+                ))));
             }
 
             if let Ok(Response::Data { partial }) = response {
                 let res = partial;
 
                 // drop remaining lines
-                while let Some(drop) = rt.block_on(self.agent.next()) {
+                while let Some(drop) = rt.block_on(self.scd.agent.next()) {
                     log::trace!("drop: {:x?}", drop);
                 }
 
@@ -249,15 +283,17 @@ impl CardClient for ScdClient {
             }
         }
 
-        Err(Error::InternalError(anyhow!("no response found")))
+        Err(Error::Smartcard(SmartcardError::Error(
+            "no response found".into(),
+        )))
     }
 
     fn init_card_caps(&mut self, caps: CardCaps) {
-        self.card_caps = Some(caps);
+        self.scd.card_caps = Some(caps);
     }
 
     fn card_caps(&self) -> Option<&CardCaps> {
-        self.card_caps.as_ref()
+        self.scd.card_caps.as_ref()
     }
 
     /// Return limit for APDU command size via scdaemon (based on Assuan
@@ -277,12 +313,12 @@ impl CardClient for ScdClient {
     }
 
     /// FIXME: not implemented yet
-    fn pinpad_verify(&mut self, _id: u8) -> Result<Vec<u8>> {
+    fn pinpad_verify(&mut self, _id: PinType) -> Result<Vec<u8>, Error> {
         unimplemented!()
     }
 
     /// FIXME: not implemented yet
-    fn pinpad_modify(&mut self, _id: u8) -> Result<Vec<u8>> {
+    fn pinpad_modify(&mut self, _id: PinType) -> Result<Vec<u8>, Error> {
         unimplemented!()
     }
 }
